<?php

namespace Service\Mongo;

use Exception;
use Service\Mongo\Types\DateType;
use Service\Mongo\Types\Type;
use Util\InstanceProvider;
use MongoDB\BSON\Regex;
use MongoDB\BSON\UTCDatetime;
use Zeedhi\Framework\DB\Mongo\Mongo;
use Zeedhi\Framework\DataSource\Configuration;

class MongoManager
{
    const PRIMARY_KEY_COLUMN = 'id';
    const MONGO_KEY_COLUMN = '_id';
    /**
     * @var Mongo
     */
    public $mongo;

    /**
     * @var Configuration
     */
    private $dataSourceConfig;

    /**
     * @var
     */
    private $data;

    /**
     * @throws Exception
     */
    public function __construct()
    {
        $this->mongo = InstanceProvider::getMongoDB();
    }

    /**
     * @param $params
     * @return array|null
     */
    private function getSearch($params): ?array
    {
        return !empty($params['search']) ? $params['search'] : null;
    }

    /**
     * @param $params
     * @return array
     */
    private function getSearchIn($params): array
    {
        return !empty($params['search_in']) ? $params['search_in'] : array();
    }

    /**
     * @return array|null
     */
    private function getFind(): ?array
    {
        return !empty($_REQUEST['find']) ? $_REQUEST['find'] : null;
    }

    /**
     * @param $params
     * @return mixed|string
     */
    private function getDynamicFilter($params)
    {
        return !empty($params['dynamic_filter']) ? $params['dynamic_filter'] : '';
    }

    /**
     * @return string|number
     */
    private function getPage($params)
    {
        return !empty($params['page']) ? $params['page'] : 1;
    }

    /**
     * @return string|number
     */
    private function getLimit($params)
    {
        return !empty($params['limit']) ? $params['limit'] : 999999;
    }

    /**
     * @return string|array
     */
    private function getOrder($params)
    {
        return !empty($params['order']) ? $params['order'] : array();
    }

    /**
     * @param $filter
     * @param $i
     * @param $lenFilter
     * @return string
     * @throws Exception
     */
    protected function getRelation($filter, $i, $lenFilter): string
    {
        $relation  = $filter[$i]['relation'];
        $relationOpMongo = '';
        if($i == 0 && $lenFilter > 1) {
            $relationOpMongo = $filter[$i + 1]['relation'] == 'OR' ? '$or' : '$and';
        } else if ($relation == 'AND' || $lenFilter == 1) {
            $relationOpMongo = '$and';
        } else if($relation == 'OR') {
            $relationOpMongo = '$or';
        } else {
            throw new Exception('relation key is not and neither or');
        }
        return $relationOpMongo;
    }

    /**
     * @throws Exception
     */
    public function findNext($collectionName, $params, $infoPagination = true, $decodeBase64 = true): array
    {
        $this->data = [];
        $mongoQuery = $this->createMongoQuery($params, $decodeBase64);
        $mongoOptions = $this->createMongoOptions($params);
        if ($infoPagination) {
            $this->data['data'] = $this->findMany($collectionName, $mongoQuery, $mongoOptions);
            $this->getInfoPagination($collectionName, $params, $decodeBase64);
            return $this->data;
        }
        $this->data = $this->findMany($collectionName, $mongoQuery, $mongoOptions);
        return $this->data;
    }

    /**
     * @param $collectionName
     * @param $params
     * @param $decodeBase64
     * @return void
     * @throws Exception
     */
    public function getInfoPagination($collectionName, $params, $decodeBase64)
    {
        $this->data["pagination"]["total"] = $this->countNext($collectionName, $params, $decodeBase64);
        $this->data["pagination"]["page"] = isset($params["page"]) ? (int)$params["page"] : 1;
        $this->data["pagination"]["limit"] = isset($params["limit"]) ? (int)$params["limit"] : 9999;
    }

    /**
     * @param $collectionName
     * @param $query
     * @param $queryOptions
     * @return array
     */
    public function findMany($collectionName, $query, $queryOptions): array
    {
        $collection = $this->mongo->getCollection($collectionName);
        $dataFound = $collection->find($query, $queryOptions);
        $rows = $dataFound->toArray();
        return $this->prepareRowsForResultSet($rows);
    }

    /**
     * @param $rows
     * @return array
     */
    protected function prepareRowsForResultSet($rows): array
    {
        $preparedRows = array();
        foreach ($rows as $row) {
            $row = $this->planningDocumentKey($row);
            $preparedRows[] = $this->filterColumnsForResultSet($row);
        }
        return $preparedRows;
    }

    /**
     * @param $row
     * @return mixed
     */
    protected function planningDocumentKey($row) {
        if(isset($row[self::MONGO_KEY_COLUMN])) {
            $row[self::MONGO_KEY_COLUMN] = Type::convertDatabaseValueToPHP($row[self::MONGO_KEY_COLUMN]);
        }
        foreach($row as &$rowProperty) {
            if(is_array($rowProperty)) {
                $rowProperty = $this->planningDocumentKey($rowProperty);
            } else if($rowProperty instanceof \MongoDB\BSON\UTCDateTime) {
                $rowProperty = Type::convertDatabaseValueToPHP($rowProperty);
                $rowProperty = $rowProperty->format('d/m/Y H:i:s');
            }
        }
        return $row;
    }

    /**
     * @param $row
     * @return array|mixed
     */
    protected function filterColumnsForResultSet($row) {
        $columnsForResultSet = array();
        if($this->dataSourceConfig){
            $columnsForResultSet = $this->dataSourceConfig->getColumnsForResultSet();
        }
        if (count($columnsForResultSet) > 0) {
            $rowFiltered = array();
            foreach ($columnsForResultSet as $columnName) {
                if (array_key_exists($columnName, $row)) {
                    $rowFiltered[$columnName] = $row[$columnName];
                }
            }
            return $rowFiltered;
        }
        return $row;
    }

    /**
     * @throws Exception
     */
    public function createMongoQuery($params, $decodeBase64 = true) {
        $search = $this->getSearch($params);
        $searchIn = $this->getSearchIn($params);
        $dynamicFilter = $this->getDynamicFilter($params);
        if ($decodeBase64) {
            $dynamicFilter = utf8_encode(base64_decode($dynamicFilter));
            $decodedDynamicFilter = empty($dynamicFilter) ? array() : json_decode($dynamicFilter, true);
        } else {
            $decodedDynamicFilter = empty($dynamicFilter) ? array() : $dynamicFilter;
        }
        $queryFilters = empty($decodedDynamicFilter) ? array() : $this->createQueryFromDynamicFilter($decodedDynamicFilter);
        return $this->createSearchInQuery($search, $searchIn, $queryFilters);
    }

    /**
     * @param $dynamicFilter
     * @return array
     * @throws Exception
     */
    public function createQueryFromDynamicFilter ($dynamicFilter): array
    {
        $finalQuery = array();
        $lenDynamicFilter = count($dynamicFilter);
        if ($lenDynamicFilter > 1) {
            $finalQuery['$and'] = array();
        }
        foreach ($dynamicFilter as $column => $filter) {
            $lenFilter = count($filter);
            $query = array();
            for ($i = 0; $i < $lenFilter; $i++) {
                if(!empty($filter[$i]['value'])) {
                    $value     = $filter[$i]['value'];
                    $operation = $filter[$i]['operation'];
                    $relationOpMongo = $this->getRelation($filter, $i, $lenFilter);
                    if (!array_key_exists($relationOpMongo, $query)) {
                        $query[$relationOpMongo] = array();
                    }
                    if (in_array($operation, array('IN', 'NOT_IN'))) {
                        if(!is_array($value)) {
                            $pieces = explode(";", $value);
                        } else {
                            $pieces = $value;
                        }
                        $query[$relationOpMongo][] = $operation == 'IN' ? array($column => array('$in' => $pieces)) : array($column => array('$not' => array('$in' => $pieces)));
                    } else if ($operation == 'BETWEEN') {
                        $betValue = is_string($value) ? explode(';', $value) : $value;
                        $query[$relationOpMongo][] = array($column =>array('$gte' => $betValue[0], '$lte' => $betValue[1]));
                    } else if (in_array($operation, array('CONTAINS', 'NOT_CONTAINS'))) {
                        $regex = new Regex($value);
                        $query[$relationOpMongo][] = $operation == 'CONTAINS' ? array($column => array('$regex' => $regex)) : array($column => array('$not' => $regex));
                    } else {
                        $mongoOperator = $this->operationSimple($operation);
                        $query[$relationOpMongo][] = array($column => array($mongoOperator => $value));
                    }
                }
            }
            if ($lenDynamicFilter > 1) {
                $finalQuery['$and'][] = $query;
            } else {
                $finalQuery = $query;
            }
        }
        return $finalQuery;
    }

    private function operationSimple($operation): string
    {
        if ($operation == 'EQUALS') {
            $mongoOperator = '$eq';
        } else if ($operation == 'NOT_EQUALS') {
            $mongoOperator = '$ne';
        } else if ($operation == 'LESS_THAN') {
            $mongoOperator = '$lt';
        } else if ($operation == 'LESS_THAN_EQUALS') {
            $mongoOperator = '$lte';
        } else if ($operation == 'GREATER_THAN') {
            $mongoOperator = '$gt';
        } else if ($operation == 'GREATER_THAN_EQUALS') {
            $mongoOperator = '$gte';
        }
        return $mongoOperator;
    }

    /**
     * @param $search
     * @param $searchIn
     * @param $queryFilters
     * @return mixed
     */
    protected function createSearchInQuery($search, $searchIn, $queryFilters) {
        if (!empty($searchIn)) {
            if (!array_key_exists('$or', $queryFilters)) {
                $queryFilters['$or'] = array();
            }
            foreach ($searchIn as $cols) {
                $regex = new Regex($search);
                $queryFilters['$or'][] = array($cols => array('$regex' => $regex));
            }
        }
        return $queryFilters;
    }

    /**
     * @param $params
     * @return array
     */
    public function createMongoOptions($params): array
    {
        $page = $this->getPage($params);
        $limit = $this->getLimit($params);
        $orderBy = $this->getOrder($params);
        $limit = (int)$limit;
        $skip = ($page - 1) * $limit;
        $order = array();
        for ($i = 0; $i < count($orderBy); $i++) {
            $content = $orderBy[$i];
            $pieces = explode('.', $content);
            $order[$pieces[0]] = $pieces[1] == 'asc' ? 1 : -1;
        }
        return array('sort' => $order, 'limit' => $limit, 'skip' => $skip);
    }

    /**
     * @param $collectionName
     * @param $params
     * @param $decodeBase64
     * @return mixed
     * @throws Exception
     */
    public function countNext($collectionName, $params, $decodeBase64 = true){
        $mongoQuery = $this->createMongoQuery($params, $decodeBase64);
        $mongoOptions = [];
        return $this->countMany($collectionName, $mongoQuery, $mongoOptions);
    }

    /**
     * @param $collectionName
     * @param $query
     * @param $queryOptions
     * @return int
     */
    public function countMany($collectionName, $query, $queryOptions): int
    {
        $collection = $this->mongo->getCollection($collectionName);
        return $collection->count($query, $queryOptions);
    }

    /**
     * @param $value
     * @return UTCDatetime
     */
    public function convertDate($value): UTCDatetime
    {
        $datetime = DateType::getDateTime($value);
        return new UTCDatetime($datetime->format('U').'000');
    }
}
